/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <cstdlib>
#include "dbconnector.h"
#include "dbexpression.h"
#include "model.h"
#include "utf8/utf8.h"


namespace morm
{

DbConnector::DbConnector()
{
	db_expression = nullptr;
	expression_allocated = false;
	logger = nullptr;
}

DbConnector::DbConnector(const DbConnector &)
{
	db_expression = nullptr;
	expression_allocated = false;
	logger = nullptr;
}


DbConnector::~DbConnector()
{
	deallocate_expression();
}


void DbConnector::set_logger(PT::Logger * logger)
{
	this->logger = logger;
}

void DbConnector::set_logger(PT::Logger & logger)
{
	this->logger = &logger;
}



void DbConnector::clear_last_query_result()
{
}


bool DbConnector::query(const PT::TextStream & stream)
{
	return false;
}


//bool DbConnector::query(const std::wstring & query)
//{
//	return false;
//}


bool DbConnector::query(const std::string & query_str)
{
	return false;
}


//bool DbConnector::query(const wchar_t * query)
//{
//	return false;
//}


bool DbConnector::query(const char * query_str)
{
	// do query
	return false;
}



bool DbConnector::query_select(const PT::TextStream & stream)
{
	return false;
}

bool DbConnector::query_update(const PT::TextStream & stream)
{
	return false;
}

bool DbConnector::query_insert(const PT::TextStream & stream)
{
	return false;
}

bool DbConnector::query_remove(const PT::TextStream & stream)
{
	return false;
}

size_t DbConnector::last_select_size()
{
	return 0;
}

std::wstring DbConnector::get_last_query_error_msg()
{
	return std::wstring();
}

bool DbConnector::was_error_in_last_query()
{
	return false;
}


void DbConnector::set_current_row_at_beginning()
{
}

void DbConnector::advance_current_row()
{
}


//void DbConnector::map_values_from_query(Model & model)
//{
//	model.clear();
//}


DbExpression * DbConnector::get_expression()
{
	allocate_default_expression_if_needed();
	return db_expression;
}


void DbConnector::generate_select_columns(PT::TextStream & stream, Model & model, const std::wstring & column_prefix)
{
	allocate_default_expression_if_needed();

	if( db_expression )
	{
		db_expression->prepare_to_new_expression();
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_FIELDS);
		db_expression->set_output_type(MORM_OUTPUT_TYPE_SELECT_COLUMNS);
		db_expression->set_column_prefix(column_prefix);
		db_expression->generate_from_model(stream, model);
	}
}


void DbConnector::generate_insert_query(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();

	if( db_expression )
	{
		db_expression->prepare_to_new_expression();
		stream << "insert into ";
		model.table_name(stream);

		stream << " (";
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_FIELDS);
		db_expression->set_output_type(MORM_OUTPUT_TYPE_DB_UPDATE);
		db_expression->generate_from_model(stream, model);

		stream << ") values (";
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_VALUES);
		db_expression->generate_from_model(stream, model);
		stream << ")";
	}
}


void DbConnector::generate_update_query(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();

	if( db_expression )
	{
		db_expression->prepare_to_new_expression();
		stream << "update ";
		model.table_name(stream);

		stream << " set ";
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_FIELDS_VALUES);
		db_expression->set_output_type(MORM_OUTPUT_TYPE_DB_UPDATE);
		db_expression->generate_from_model(stream, model);

		stream << " where ";
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_FIELDS_VALUES);
		db_expression->set_output_type(MORM_OUTPUT_TYPE_DB_PRIMARY_KEY);
		db_expression->generate_from_model(stream, model);
	}
}


void DbConnector::generate_remove_query(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();

	if( db_expression )
	{
		db_expression->prepare_to_new_expression();
		stream << "delete from ";
		model.table_name(stream);
		stream << " where ";
		db_expression->set_work_mode(MORM_WORK_MODE_MODEL_FIELDS_VALUES);
		db_expression->set_output_type(MORM_OUTPUT_TYPE_DB_PRIMARY_KEY);
		db_expression->generate_from_model(stream, model);
	}
}

bool DbConnector::insert(PT::TextStream & stream, Model & model)
{
	generate_insert_query(stream, model);
	return query_insert(stream);
}


bool DbConnector::update(PT::TextStream & stream, Model & model)
{
	generate_update_query(stream, model);
	return query_update(stream);
}


bool DbConnector::remove(PT::TextStream & stream, Model & model)
{
	generate_remove_query(stream, model);
	return query_remove(stream);
}


void DbConnector::deallocate_expression()
{
	if( expression_allocated )
	{
		delete db_expression;
		db_expression = nullptr;
		expression_allocated = false;
	}
}



void DbConnector::allocate_default_expression_if_needed()
{
	if( !db_expression )
	{
		allocate_default_expression();
	}
}

const char * DbConnector::get_field_string_value(const char * field_name)
{
	return nullptr;
}

const char * DbConnector::get_field_string_value(const wchar_t * field_name)
{
	return nullptr;
}






void DbConnector::get_value(const char * value_str, char & field_value)
{
	field_value = *value_str;
	value_str += 1;

	if( *value_str != 0 )
	{
		// value has more than one charater, put some error?
	}
}



void DbConnector::get_value(const char * value_str, unsigned char & field_value)
{
	field_value = *(const unsigned char*)value_str;
	value_str += 1;

	if( *value_str != 0 )
	{
		// value has more than one charater, put some error?
	}
}


void DbConnector::get_value(const char * value_str, std::wstring & field_value)
{
	// CHECKME
	// what about \0 in val_str?
	// it is escaped somehow?
	PT::UTF8ToWide(value_str, field_value);
}


void DbConnector::get_value(const char * value_str, std::string & field_value)
{
	field_value = value_str;
}


void DbConnector::get_value(const char * value_str, bool & field_value)
{
	// IMPROVE ME
	// this 't' is locale dependent
	field_value = (value_str[0]=='t' || value_str[0]=='y' || value_str[0]=='1');
}


void DbConnector::get_value(const char * value_str, short & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = (short)PT::Toi(value_str, 10);
}


void DbConnector::get_value(const char * value_str, unsigned short & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = (unsigned short)PT::Toui(value_str, 10);
}


void DbConnector::get_value(const char * value_str, int & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Toi(value_str, 10);
}


void DbConnector::get_value(const char * value_str, unsigned int & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Toui(value_str, 10);
}


void DbConnector::get_value(const char * value_str, long & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Tol(value_str, 10);
}


void DbConnector::get_value(const char * value_str, unsigned long & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Toul(value_str, 10);
}


void DbConnector::get_value(const char * value_str, long long & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Toll(value_str, 10);
}


void DbConnector::get_value(const char * value_str, unsigned long long & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = PT::Toull(value_str, 10);
}


void DbConnector::get_value(const char * value_str, float & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = strtof(value_str, 0);
}


void DbConnector::get_value(const char * value_str, double & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = strtod(value_str, 0);
}


void DbConnector::get_value(const char * value_str, long double & field_value)
{
	// IMPROVE ME give some overflow checking
	field_value = strtold(value_str, 0);
}


void DbConnector::get_value(const char * value_str, PT::Date & field_value)
{
	// IMPROVE ME give some log if parsing failed
	field_value.Parse(value_str);
}


const char * DbConnector::query_last_sequence(const wchar_t * sequence_table_name)
{
	return nullptr;
}

}
