/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef headerfile_morm_finder
#define headerfile_morm_finder

#include <vector>
#include <list>
#include <set>
#include "model.h"
#include "textstream/textstream.h"
#include "dbconnector.h"
#include "modelconnector.h"


namespace morm
{

template<typename ModelClass>
class Finder
{

public:


	Finder()
	{
		model_connector = nullptr;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}


	Finder(PT::TextStream & out_stream)
	{
		this->out_stream = &out_stream;
		this->model_connector = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(ModelConnector & model_connector)
	{
		this->model_connector = &model_connector;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(ModelConnector * model_connector)
	{
		this->model_connector = model_connector;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(PT::TextStream & out_stream, ModelConnector & model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = &model_connector;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(PT::TextStream & out_stream, ModelConnector * model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = model_connector;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}


	bool was_error()
	{
		return was_query_error;
	}

	std::wstring get_error_msg()
	{
		return last_query_error;
	}


	Finder<ModelClass> & prefix(const wchar_t * prefix)
	{
		column_prefix = prefix;
		return *this;
	}


	Finder<ModelClass> & prepare_to_select()
	{
		was_query_error = false;
		last_query_error.clear();
		model_data = nullptr;
		column_prefix.clear();

		if( model_connector )
		{
			if( !out_stream )
			{
				out_stream = model_connector->get_stream();
			}

			if( out_stream )
			{
				set_db_expression();

				out_stream->clear();
				model.set_connector(model_connector);
			}
			else
			{
				was_query_error = true;
				last_query_error = L"out stream object is required";
			}
		}
		else
		{
			was_query_error = true;
			last_query_error = L"model connector object is required";
		}

		return *this;
	}


	Finder<ModelClass> & select(ModelConnector & model_connector)
	{
		this->model_connector = &model_connector;
		return select();
	}

	Finder<ModelClass> & select(ModelConnector * model_connector)
	{
		this->model_connector = model_connector;
		return select();
	}

	Finder<ModelClass> & select(PT::TextStream & out_stream, ModelConnector & model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = &model_connector;
		return select();
	}

	Finder<ModelClass> & select(PT::TextStream & out_stream, ModelConnector * model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = model_connector;
		return select();
	}


	Finder<ModelClass> & select(bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		if( model_connector && out_stream )
		{
			(*out_stream) << "SELECT ";
			model.generate_select_columns(*out_stream, column_prefix);
			(*out_stream) << " FROM ";
			model.table_name(*out_stream);

			if( !column_prefix.empty() )
				(*out_stream) << " AS " << column_prefix; // what about escaping?

			(*out_stream) << " ";
		}

		return *this;
	}

	Finder<ModelClass> & select(Model::ModelData * model_data, bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		this->model_data = model_data;
		return select(false);
	}

	Finder<ModelClass> & select(Model::ModelData & model_data, bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		this->model_data = &model_data;
		return select(false);
	}

	Finder<ModelClass> & where()
	{
		if( out_stream && db_expression )
		{
			(*out_stream) << " WHERE ";
			db_expression->prepare_to_where_clause();
		}

		return *this;
	}


	Finder<ModelClass> & group_or()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_or(*out_stream);
		}

		return *this;
	}


	Finder<ModelClass> & group_and()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_and(*out_stream);
		}

		return *this;
	}


	Finder<ModelClass> & group_end()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_end(*out_stream);
		}

		return *this;
	}




	template<typename FieldValue>
	Finder<ModelClass> & eq(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_EQ);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & lt(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_LT);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & gt(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_GT);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & le(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_LE);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & ge(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_GE);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false);
		}

		return *this;
	}



	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::set<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::list<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::vector<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	Finder<ModelClass> & page(size_t page_number, size_t page_size)
	{
		if( out_stream && db_expression )
		{
			db_expression->page(*out_stream, page_number, page_size);
		}

		return *this;
	}


	Finder<ModelClass> & raw(const char * sql, bool add_spaces = true)
	{
		if( out_stream )
		{
			if( add_spaces )
				(*out_stream) << ' ';

			(*out_stream) << sql;

			if( add_spaces )
				(*out_stream) << ' ';
		}

		return *this;
	}


	Finder<ModelClass> & raw(long val, bool add_spaces = true)
	{
		if( out_stream )
		{
			if( add_spaces )
				(*out_stream) << ' ';

			(*out_stream) << val;

			if( add_spaces )
				(*out_stream) << ' ';
		}

		return *this;
	}


	bool get(ModelClass & result)
	{
		bool res = false;
		result.set_connector(model_connector);
		result.clear();

		if( model_connector && out_stream )
		{
			DbConnector * db_connector = model_connector->get_db_connector();

			if( db_connector )
			{
				try
				{
					result.model_data = model_data;
					result.before_select();
					res = db_connector->query_select(*out_stream);

					if( res )
					{
						if( db_connector->last_select_size() == 1 )
						{
							result.set_save_mode(Model::DO_UPDATE_ON_SAVE); // IMPROVE ME check if there is a primary key
							result.map_values_from_query();
							result.after_select();
						}
						else
						{
							res = false;
							// log some error or throw an exception if there are more items than one?
						}
					}
					else
					{
						result.after_select_failure();
					}
				}
				catch(...)
				{
					res = false;
					// throw something?
					// if yes then make sure to call db_connector->clear_last_query_result();
				}

				result.model_data = nullptr;
				db_connector->clear_last_query_result();
			}
		}

		return res;
	}


	ModelClass get()
	{
		get(model);
		return model;
	}


	bool get_list(std::list<ModelClass> & result, bool clear_list = true)
	{
		bool res = false;

		if( clear_list )
		{
			result.clear();
		}

		if( model_connector && out_stream )
		{
			DbConnector * db_connector = model_connector->get_db_connector();

			if( db_connector )
			{
				res = db_connector->query_select(*out_stream);

				try
				{
					if( res )
					{
						size_t len = db_connector->last_select_size();
						db_connector->set_current_row_at_beginning();

						for(size_t i = 0 ; i < len ; ++i)
						{
							result.emplace_back(); // it returns a reference from c++17
							ModelClass & added_model = result.back();

							try
							{
								added_model.set_connector(model_connector);
								added_model.clear();
								added_model.set_save_mode(Model::DO_UPDATE_ON_SAVE); // IMPROVE ME check if there is a primary key
								added_model.model_data = model_data;
								added_model.before_select();
								added_model.map_values_from_query();
								added_model.after_select();
							}
							catch(...)
							{
								res = false;
								// throw or something?
								// make sure to call db_connector->clear_last_query_result()
							}

							added_model.model_data = nullptr;
							db_connector->advance_current_row();
						}

						// if there was a failure we do not call after_select_failure()
						// because there are no any objects in the list
					}
					else
					{
						// log
					}
				}
				catch(...)
				{
					res = false;
					// throw or something?
					// make sure to call db_connector->clear_last_query_result()
				}

				db_connector->clear_last_query_result();
			}
		}

		return res;
	}

	std::list<ModelClass> get_list()
	{
		std::list<ModelClass> result;
		get_list(result, false);
		return std::move(result);
	}


private:


	PT::TextStream * out_stream;
	ModelConnector * model_connector;
	DbExpression * db_expression;
	ModelClass model;
	bool was_query_error;
	std::wstring last_query_error;
	Model::ModelData * model_data;
	std::wstring column_prefix;



	void set_db_expression()
	{
		if( model_connector )
		{
			DbConnector * db_connector = model_connector->get_db_connector();

			if( db_connector )
			{
				db_expression = db_connector->get_expression();
			}
		}
	}


};

} // namespace

#endif
